
#pragma once

#include <cassert>
#include <cstdint>
#include <climits>
#include <string>
#include <type_traits>
#include <memory>
#include <WS2tcpip.h>
#include "dosbox.h"

#ifndef LENGTH_OF
    #define LENGTH_OF(a) (sizeof(a) / sizeof(*a))
#endif

class TCPSocket
{
    TCPSocket(TCPSocket&);
    TCPSocket& operator=(TCPSocket&);
    
    SOCKET      _socket;
    std::string _error;
    
    public:
        TCPSocket(char const* const ip, uint32_t const port) : _socket(INVALID_SOCKET)
        {
            WSADATA wsaData = {0};
            if (auto const err = WSAStartup(MAKEWORD(2, 2), &wsaData)) {
                char buffer[0x1ff];
                sprintf_s(buffer, "WSA failed to init. Err Code: 0x%x", err);
                _error = buffer;
                return;
            }
            char portStr[32]; sprintf_s(portStr, "%u", port);
            
            addrinfo* result = nullptr;
            addrinfo  specs  = {0};
            specs.ai_flags    = AI_NUMERICHOST;
            specs.ai_family   = AF_UNSPEC;
            specs.ai_socktype = SOCK_STREAM;
            specs.ai_protocol = IPPROTO_TCP;
            
            if (auto const err = getaddrinfo(ip, portStr, &specs, &result)) {
                char buffer[0x1ff];
                sprintf_s(buffer, "Invalid ip/port: %s:%s. Err Code - 0x%x08; Err Msg - %s"
                    , ip, portStr, err, gai_strerror(err));
                _error = buffer;
                return;
            }
            
            _socket = socket(result->ai_family, result->ai_socktype, result->ai_protocol);
            if (_socket == INVALID_SOCKET) {
                char buffer[0x1ff];
                sprintf_s(buffer, "Failed to create socket: Err Code - 0x%08x.", WSAGetLastError());
                _error = buffer;
                freeaddrinfo(result);
                return;
            }
            
            auto const errConnection = connect(_socket, result->ai_addr, (int)result->ai_addrlen);
            freeaddrinfo(result); result = nullptr;
            
            if (errConnection) {
                char buffer[0x1ff];
                sprintf_s(buffer, "Failed to connect: Err Code - 0x%08x.", WSAGetLastError());
                _error = buffer;

                closesocket(_socket);
                _socket = INVALID_SOCKET;
            }
        }
        ~TCPSocket()
        {
            closesocket(_socket);
            WSACleanup();
        }
        
        bool               IsValid () const { return _socket != INVALID_SOCKET; }
        std::string const& ErrorGet() const { return _error; }
        
        TCPSocket& Write(char const datum)
        {
            assert(IsValid());
            send(_socket, &datum, sizeof(datum), 0);
            return *this;
        }
        
        // Only use WriteRaw if the data is a stream of oclets
        // or you are 200% certain the other end has the same byte ordering as you.
        template<typename T>
        typename std::enable_if<  std::is_pod    <T>::value  &&
                                (!std::is_array  <T>::value) &&
                                (!std::is_pointer<T>::value), TCPSocket&>::type
        WriteRaw(T const& datum)
        {
            assert(IsValid());
            send(_socket, reinterpret_cast<char const*>(&datum), sizeof(datum), 0);
            return *this;
        }
        
        template<typename T>
        TCPSocket& WriteRaw(T const* const data
                           ,typename std::enable_if<  std::is_pod    <T>::value  &&
                                                    (!std::is_pointer<T>::value) &&
                                                    (sizeof(T) == 1            ), size_t>::type
                            const size)
        {
            assert(IsValid());
            auto const ret = send(_socket, reinterpret_cast<char const*>(data), size, 0);
            return *this;
        }
};

